from __future__ import annotations

import glob
import itertools
import math
import pathlib
from typing import *

import lunas.dataset.core as core

__all__ = ['Array', 'Range', 'Zip', 'Glob']


class Array(core.Dataset, core.Sizable):
    """A dataset that wraps around any data as an iterable array.

    This dataset accepts any data. Note that the iterable data will be converted to an object with give array_type.
    """

    def __init__(self, data: Any, array_type: TypeVar('T') = list, name: str = None):
        super().__init__(name)
        self._data = array_type(data)
        self._length = None

    @property
    def length(self):
        if self._length is None:
            self._length = len(self._data)
        return self._length

    def generator(self):
        for x in self._data:
            yield x


class Range(core.Dataset, core.Sizable):
    """Range dataset.

    This dataset simulates the builtin `range` function.
    """

    def __init__(self, start: int, stop: int = None, step: int = None, name: str = None):
        super().__init__(name)
        if stop is None:
            stop = start
            start = 0

        step = step or 1

        self._start = start
        self._stop = stop
        self._step = step

        self._length = None

    @property
    def length(self):
        if self._length is None:
            start, stop, step = self._start, self._stop, self._step
            self._length = max(0, int(math.ceil((stop - start) / step)))
        return self._length

    def generator(self):
        for x in range(self._start, self._stop, self._step):
            yield x


class Zip(core.NestedN, core.Sizable):
    """Zip multiple dataset.

    This dataset zips multiple datasets, potentially with different sizes.
    """

    def __init__(self, datasets: Iterable[core.Dataset], mode: str = '=', padding: bool = False,
                 name: str = None):
        """Initialises the dataset.

        Args:
            datasets: the datasets to zip.
            mode: a character, available options include '=' '<' and '>'.
                '=' requires the datasets to have the same sizes;
                '<' behaves similarly to the builtin `zip`, which truncate the bigger datasets to align with the
                smallest one;
                '>' is similar to `itertools.zip_longest`, fill the smaller datasets with strategy specified by
                `padding`.
            padding: a boolean value that determines how to pad the small datasets when they are exhausted.
                A `False` will produce `None` as padding, while `True` will continue producing elements from smaller
                datasets. Only works when mode is '>'.
            name: name of the dataset.
        """
        super().__init__(datasets, name)
        if mode not in ('=', '<', '>'):
            raise ValueError(f'Unknown mode: {mode}')

        self._length = None

        if mode == '=':
            sizes = tuple(map(len, datasets))
            if len(set(sizes)) > 1:
                raise RuntimeError(f'Datasets must have exactly the same sizes ({tuple(sizes)}).')

            self._length = sizes[0]

        if not isinstance(padding, bool):
            raise ValueError(f'padding ({padding}) must be a bool value.')

        self._mode = mode
        self._padding = padding

    @property
    def length(self):
        if self._length is None:
            sizes = tuple(map(len, self._datasets))
            if self._mode == '<':
                self._length = min(sizes)
            elif self._mode == '>':
                self._length = max(sizes)
            else:
                raise ValueError(f'Unknown mode: {self._mode}')
        return self._length

    def generator(self):
        if self._mode in ['=', '<']:
            for x in zip(*self._datasets):
                yield x
        else:
            if not self._padding:
                it = itertools.zip_longest(*self._datasets)
                for x in it:
                    yield x
            else:
                # DO NOT USE itertools.cycle EVER!
                # Since itertools.cycle actually stores all elements after one-time iteration,
                # this is super-memory-consuming and breaks the internal state maintenance of a `Dataset`.
                datasets = [itertools.chain.from_iterable(itertools.repeat(d))
                            if len(d) < len(self) else d
                            for d in self._datasets]
                # Additionally, islice is used here. zip will stop when one iterator raises StopIteration,
                # so any datasets before it will unexpectedly advance one element.
                for x in itertools.islice(zip(*datasets), len(self)):
                    yield x


class Glob(core.Dataset, core.Sizable):
    """Glob dataset.

    This dataset uses standard glob module to wrap matched directories/files for given pattern into a dataset.
    """

    def __init__(self, pattern: str, recursive: bool = False, expand_user: bool = True, name: str = None):
        """Initialises the dataset.

        Args:
            pattern: a glob pattern.
            recursive: whether matches recursively.
            expand_user: whether expands the user home path.
            name: name of the dataset.
        """
        super().__init__(name)
        self._pattern = str(pathlib.Path(pattern).expanduser() if expand_user else pathlib.Path(pattern))
        self._files = sorted(glob.glob(self._pattern, recursive=recursive))
        self._length = len(self._files)

    @property
    def length(self):
        return self._length

    def generator(self):
        for x in self._files:
            yield x



""" Abstract dataset classes and implementation of core datasets.
"""

from __future__ import annotations

import abc
import itertools
import math
import sys
from typing import *

import numpy

__all__ = [
    'Dataset',
    'Sizable',
    'NestedN',
    'Nested',
    'NestedSizable',
    'Map',
    'Where',
    'Repeat',
    'Interleave',
    'Shuffle',
    'Sort',
    'Slice',
    'Shard',
    'Enumerate',
    'Group',
    'Concat'
]


class Dataset(abc.ABC):
    """An abstract class representing a dataset.

    A dataset is a wrapper for certain data source, providing interfaces to process and filter
    dataset elements during iteration, which are evaluated at runtime.

    The elements can be accessed through Python's iterator interface.
    """

    def __init__(self, name: str = None):
        """Initialises the dataset.
        Args:
            name: name of the dataset.
        """
        super().__init__()
        self._ptr = 0
        self._ref_count = 0
        self._resumed = False
        self._resumable = True

        self._name = name or self.__class__.__name__

    @abc.abstractmethod
    def generator(self):
        """A generator generates samples from the data source."""
        raise NotImplementedError

    def __iter__(self):
        """Returns an iterator of the dataset.

        Optionally, the iteration can be resumed by loading previously saved state, see `Dataset.state` and `Dataset.load`.
        """
        if not self._resumed:
            self._reset()

        it = self.generator()
        if self._resumable and self._ptr > 0:
            it = itertools.islice(it, self._ptr, None)
        self._resumed = False

        for x in it:
            self._ptr += 1
            if x is not None:
                yield x

        self._reset()

    @property
    def name(self):
        return self._name

    def state(self) -> dict:
        """Returns a checkpoint state

        The checkpoint can be used to resume iteration from previous stopping position.

        Returns:
            A dictionary containing necessary information to recover iteration state.
        """
        if not self._resumable:
            raise RuntimeError(f'dataset ({self.name}) is not resumable.')
        return {'ptr': self._ptr, 'name': self.name}

    def load(self, state: dict) -> None:
        """Recovers from a checkpoint state.

        Once recovered, we can create an iterator for the dataset to continue iteration from previous iteration.

        Args:
            state: a dictionary containing necessary information to recover iteration state.
        """
        if state is None:
            return

        if state['name'] != self.name:
            raise ValueError(f'name in state ({state["name"]}) should be the same as self.name ({self.name})')

        self._ptr = state['ptr']
        self._resumed = True

    def _chk_and_inc_ref(self) -> None:
        """Ensures a dataset to be referenced by at most 1 dataset."""
        if self._ref_count > 0:
            raise RuntimeError(f'Dataset `{self}` is only allowed to be wrapped in another dataset once, '
                               'you should consider create a new instance instead.')
        else:
            self._ref_count += 1

    def _reset(self) -> None:
        """Resets iteration state."""
        self._ptr = 0
        self._resumed = False

    def map(self, fn: Callable[[Any], Any], unpack_args: bool = False, unpack_kwargs: bool = False,
            name: str = None) -> Map:
        """See `Map` class.

        Returns:
            A `Map` dataset.
        """
        return Map(self, fn, unpack_args, unpack_kwargs, name)

    def where(self, predicate: Callable[[Any], bool], unpack_args: bool = False, unpack_kwargs: bool = False,
              name: str = None) -> Where:
        """See `Where` class.

        Returns:
            A `Where` dataset.
        """
        return Where(self, predicate, unpack_args, unpack_kwargs, name)

    def repeat(self, n: int = None, name: str = None) -> Repeat:
        """See `Repeat` class.

        Returns:
            A `Repeat` dataset.
        """
        return Repeat(self, n, name)

    def shard(self, num_shards: int, shard_index: int, name: str = None) -> Shard:
        """See `Shard` class.

        Returns:
            A `Shard` dataset.
        """
        return Shard(self, num_shards, shard_index, name)

    def shuffle(self, buffer_size: int, name: str = None) -> Shuffle:
        """See `Shuffle` class.

        Returns:
            A `Shuffle` dataset.
        """
        return Shuffle(self, buffer_size, name)

    def sort(self, buffer_size: int, key: Callable[[Any], Any] = None, name: str = None) -> Sort:
        """See `Sort` class.

        Returns:
            A `Sort` dataset.
        """
        return Sort(self, buffer_size, key, name)

    def slice(self, start: int = None, stop: int = None, step: int = None, name: str = None) -> Slice:
        """See `Slice` class.

        Returns:
            A `Slice` dataset.
        """
        return Slice(self, start, stop, step, name)

    def take(self, n: int, name: str = None) -> Slice:
        """See `Take` class.

        Returns:
            A `Take` dataset.
        """
        return Slice(self, stop=n, name=name)

    def skip(self, n: int, name: str = None) -> Slice:
        """See `Skip` class.

        Returns:
            A `Skip` dataset.
        """
        return Slice(self, start=n, name=name)

    def enumerate(self, start: int = 0, name: str = None) -> Enumerate:
        """See `Enumerate` class.

        Returns:
            A `Enumerate` dataset.
        """
        return Enumerate(self, start, name)

    def group(self, group_size: int, name: str = None) -> Group:
        """See `Group` class.

        Returns:
            A `Group` dataset.
        """
        return Group(self, group_size, name)

    def flatten(self, name: str = None) -> Flatten:
        """See `Flatten` class.

        Returns:
            A `Flatten` dataset.
        """
        return Flatten(self, name)

    def concat(self, other: 'Dataset', name: str = None) -> Concat:
        """See `Concat` class.

        Returns:
            A `Concat` dataset.
        """
        return Concat([self, other], name)


class Sizable(abc.ABC):
    """An interface to evaluate length of a dataset."""

    @property
    def length(self):
        raise NotImplementedError

    def __len__(self):
        """Returns size of the dataset."""
        return self.length


class NestedN(Dataset, abc.ABC):
    """A wrapper around multiple datasets."""

    def __init__(self, datasets: Iterable[Dataset], name: str = None):
        super().__init__(name)
        if not isinstance(datasets, (tuple, list)):
            raise TypeError(f'datasets ({type(datasets)}) must be a tuple or a list.')
        datasets = tuple(datasets)
        for dataset in datasets:
            if not isinstance(dataset, Dataset):
                raise TypeError(f'datasets ({[type(dataset) for dataset in datasets]}) must inherit Dataset.')
        for x in datasets:
            x._chk_and_inc_ref()

        self._datasets: tuple = datasets
        self._resumable = all(dataset._resumable for dataset in datasets)

    @abc.abstractmethod
    def generator(self):
        """See base class."""
        raise NotImplementedError

    def _reset(self) -> None:
        super()._reset()
        for x in self._datasets:
            x._reset()


class Nested(NestedN, abc.ABC):
    """A wrapper around a single dataset."""

    def __init__(self, dataset: Dataset, name: str = None):
        super().__init__([dataset], name)
        self._dataset = dataset


class NestedSizable(Nested, Sizable, abc.ABC):
    """A sizable nested dataset."""

    def __init__(self, dataset: Dataset, name: str = None):
        super().__init__(dataset, name)
        self._length = None

    @property
    def length(self):
        if self._length is None:
            self._length = len(self._dataset)
        return self._length


class Map(NestedSizable):
    """Transform a sample.

    This dataset applies transformation to every sample in the dataset. Note that the transformation won't be applied until
    we try to access an element through an iterator.

    Example usage:

        ds = Range(100)
        ds = ds.map(lambda x: x * 2)
                .map(lambda x: x + 1)
        transformed = list(ds)
    """

    def __init__(self, dataset: Dataset, fn: Callable[[Any], Any], unpack_args: bool = False,
                 unpack_kwargs: bool = False, name: str = None) -> None:
        """Initialises the dataset.

        Args:
            dataset: the dataset to apply transformation.
            fn: a function that accepts a sample as input and returns a transformed output.
            unpack_args:
            unpack_kwargs:
            name:
        """
        super().__init__(dataset, name)
        if not callable(fn):
            raise ValueError(f'fn ({type(fn)}) must be callable.')
        if unpack_args and unpack_kwargs:
            raise ValueError(f'unpacking as args and kwargs are conflicting.')

        self._fn = fn
        self._unpack_args = unpack_args
        self._unpack_kwargs = unpack_kwargs

    def generator(self):
        if self._unpack_args:
            for x in self._dataset:
                yield self._fn(*x)
        elif self._unpack_kwargs:
            for x in self._dataset:
                yield self._fn(**x)
        else:
            for x in self._dataset:
                yield self._fn(x)


class Where(Nested):
    """Filter a sample by given predicate.

    This dataset applies filter to every sample in the dataset. Note that the filter won't be applied until we try to access
    an element through the iterator. A sample is only preserved if the predicate is evaluated True.

    Example usage:

        ds = Range(100)
        ds = ds.map(lambda x: x * 2)
                .where(lambda x: x % 10 == 0)
        transformed = list(ds)
    """

    def __init__(self, dataset: Dataset, predicate: Callable[[Any], bool], unpack_args: bool = False,
                 unpack_kwargs: bool = False, name: str = None):
        """
        Args:
            dataset: the dataset to apply filter.
            predicate: a function that accepts a sample as input and returns a `bool` value to indicate whether
            this sample should be preserved.
            unpack_args:
            unpack_kwargs:
            name:
        """
        super().__init__(dataset, name)
        if not callable(predicate):
            raise ValueError(f'predicate ({type(predicate)}) must be callable.')
        if unpack_args and unpack_kwargs:
            raise ValueError(f'unpacking as args and kwargs are conflicting.')

        self._predicate = predicate
        self._unpack_args = unpack_args
        self._unpack_kwargs = unpack_kwargs

    def generator(self):
        if self._unpack_args:
            for x in self._dataset:
                if self._predicate(*x):
                    yield x
        elif self._unpack_kwargs:
            for x in self._dataset:
                if self._predicate(**x):
                    yield x
        else:
            for x in self._dataset:
                if self._predicate(x):
                    yield x


class Repeat(NestedSizable):
    """Repeat a dataset."""

    def __init__(self, dataset: Dataset, n: int = None, name: str = None):
        """Initialises the dataset.
        Args:
            dataset: a dataset to repeat.
            n: a value indicates how many times the dataset will repeat. Repeats endlessly if n is None.
            name:
        """
        super().__init__(dataset, name)
        if not (n is None or n > 0):
            raise ValueError(f'n ({n}) must be a positive integer or None.')

        self._n = n

    @property
    def length(self):
        if self._length is None:
            self._length = len(self._dataset) * self._n if self._n is not None else sys.maxsize
        return self._length

    def generator(self):
        repeats = itertools.repeat(self._dataset, self._n) if self._n else itertools.repeat(self._dataset)
        for x in itertools.chain.from_iterable(repeats):
            yield x


class Interleave(Nested):
    """Interleave iteration between multiple datasets.

    map_fn maps each element in dataset to a new dataset, then we interleave iteration according to cycle_length and
    block_length. By cycle_length, we iterate over the first cycle_length datasets alternatively until any of which
    is exhausted and then switch to the next available dataset. At each cycle, we sequentially consumed block_length
    examples from current dataset.

    One use case of this class, is partitioning a large dataset into multiple directories and shuffling each subset
    within the directory independently, then interleave access to different subsets to approximate global shuffling
    of the large dataset without re-shuffling the whole dataset, which can be rather time-consuming.

    For example:
        directories = Glob('path/to/dataset')
        ds = InterleaveDataset(directories, lambda filename: TextLine(filename), 3, 1000)
    """

    def __init__(self, dataset: Dataset, map_fn: Callable[[Any], Dataset], cycle_length: int, block_length: int,
                 name: str = None):
        """Initialises the dataset.

        Args:
            dataset: a dataset object to generate subsets.
            map_fn: mapping element of `dataset` to a new dataset.
            cycle_length: number of consecutive datasets to access.
            block_length: number of consecutive elements to access.
            name: name of the dataset.
        """
        super().__init__(dataset.map(lambda x: iter(map_fn(x))), name)
        if not callable(map_fn):
            raise ValueError(f'map_fn ({type(map_fn)}) must be callable.')
        if not (cycle_length > 0):
            raise ValueError(f'cycle_length ({cycle_length}) must be a positive integer.')
        if not (block_length > 0):
            raise ValueError(f'block_length ({block_length}) must be a positive integer.')

        self._cycle_length = cycle_length
        self._block_length = block_length

    def generator(self):
        it = iter(self._dataset)
        xs = [x for x in itertools.islice(it, self._cycle_length)]
        i = 0
        cycle_length = len(xs)
        while cycle_length > 0:
            i = i % cycle_length
            x = xs[i]
            items = list(itertools.islice(x, self._block_length))
            if not items:
                xs.pop(i)
                try:
                    xs.append(next(it))
                except StopIteration:
                    i -= 1
                    cycle_length -= 1
            else:
                for item in items:
                    yield item
                i += 1


class Shuffle(NestedSizable):
    """Shuffle a dataset.

    This dataset provides a queue-based shuffling to shuffle a dataset without loading all data into memory at once.
    The algorithm is an approximation of global shuffling.
    """

    def __init__(self, dataset: Dataset, buffer_size: int, name: str = None):
        """Initialises the dataset.

        Args:
            dataset: a dataset to shuffle.
            buffer_size: shuffles within the given buffer size.
            name:
        """
        super().__init__(dataset, name)
        if not (buffer_size > 1):
            raise ValueError(f'buffer_size ({buffer_size}) must be a positive integer greater than 1.')

        self._buffer_size = buffer_size

    def generator(self):
        it = iter(self._dataset)
        buffer = list(itertools.islice(it, self._buffer_size))
        buf_size = len(buffer)
        indices = numpy.nditer(numpy.random.randint(0, buf_size, buf_size))

        for x in it:
            try:
                i = next(indices)
            except StopIteration:
                indices = numpy.nditer(numpy.random.randint(0, buf_size, buf_size))
                i = next(indices)
            yield buffer[i]
            buffer[i] = x

        numpy.random.shuffle(buffer)

        for x in buffer:
            yield x


class Sort(NestedSizable):
    """Sort a dataset.

    This dataset provides a partial sorting to sort elements from a dataset in a custom-sized buffer, so that we don't
    load all data into memory at once. It is typically used to gather sequences of similar length together to reduce
    redundant computational overheads, while preserving randomness introduced by shuffling. To that end, we must use
    a larger buffer_size in Shuffle than Sort.

    For example:
        ds = TextLine(filename)
        ds = ds.shuffle(buffer_size=100000).sort(buffer_size=1000, key=lambda x: len(x.split()))
    """

    def __init__(self, dataset: Dataset, buffer_size: int, key: Callable[[Any], Any] = None, name: str = None):
        """Initialises the dataset.

        Args:
            dataset: a dataset to sort.
            buffer_size: sorts within the given buffer size.
            key: a function to get key for comparison.
            name:
        """
        super().__init__(dataset, name)
        if not (buffer_size > 1):
            raise ValueError(f'buffer_size ({buffer_size}) must be a positive integer greater than 1.')
        if not (key is None or callable(key)):
            raise ValueError(f'key ({type(key)}) must be callable.')

        self._buffer_size = buffer_size
        self._key = key

    def generator(self):
        it = iter(self._dataset)
        buf_size = self._buffer_size

        while True:
            buffer = sorted(itertools.islice(it, buf_size), key=self._key)

            for x in buffer:
                yield x

            if not buffer:
                break


class Slice(NestedSizable):
    """Slice a dataset."""

    def __init__(self, dataset: Dataset, start: int = 0, stop: int = None, step: int = 1, name: str = None):
        """Initialises the dataset

        Args:
            dataset: a dataset to slice.
            start: starting position.
            stop: stopping position.
            name:
        """
        super().__init__(dataset, name)
        if start is not None and not (start >= 0):
            raise ValueError(f'start ({start}) for Slice must be a non-negative integer or None.')
        if stop is not None and not (stop >= 0):
            raise ValueError(f'stop ({stop}) for Slice must be a non-negative integer or None.')
        if step is not None and not (step > 0):
            raise ValueError(f'step ({step}) for Slice must be a positive integer or None.')

        start = start or 0
        step = step or 1

        self._start = start
        self._stop = stop
        self._step = step

    @property
    def length(self):
        if self._length is None:
            self._length = max(0, math.ceil((len(self._dataset) - self._start) / self._step))
        return self._length

    def generator(self):
        for x in itertools.islice(self._dataset, self._start, self._stop, self._step):
            yield x


class Shard(Slice):
    """Shard the dataset.

    This dataset partitions elements alternately into several exclusive partitions and only one specific partition
    is loaded. This is typically used in distributed settings, where we want to ensure our training
    workers process different parts of the training data.

    For example, in worker-0:
        ds = TextLine(filename)
        ds = ds.shard(2, 0)

    And worker-1:
        ds = TextLine(filename)
        ds = ds.shard(2, 1)
    """

    def __init__(self, dataset: Dataset, num_shards: int, shard_index: int, name: str = None):
        """Initialises the dataset.
        Args:
            dataset: a dataset object to shard.
            num_shards: number of total shards.
            shard_index: index of current shard.
            name:
        """
        if not (shard_index >= 0):
            raise ValueError(f'shard_index ({shard_index}) must be an non-negative integer.')
        if not (shard_index < num_shards):
            raise ValueError(f'shard_index ({shard_index}) must be smaller than num_shards ({num_shards}).')

        super().__init__(dataset, shard_index, None, num_shards, name)


class Enumerate(NestedSizable):
    """Enumerate a dataset.

    This dataset simulates the builtin `enumerate` and attach an index to each element in the given dataset.
    """

    def __init__(self, dataset: Dataset, start: int = 0, name: str = None):
        super().__init__(dataset, name)
        self._start = start

    def generator(self):
        for x in enumerate(self._dataset, self._start):
            yield x


class Group(NestedSizable):
    """Group consecutive elements into an element."""

    def __init__(self, dataset: Dataset, group_size: int, name: str = None):
        if not (group_size >= 1):
            raise ValueError(f'group_size ({group_size}) must be a positive integer greater than 1.')

        super().__init__(dataset, name)

        self._group_size = group_size

    @property
    def length(self):
        if self._length is None:
            self._length = math.ceil(len(self._dataset) / self._group_size)
        return self._length

    def generator(self):
        it = iter(self._dataset)
        stopped = False
        while not stopped:
            group = []
            for _ in range(self._group_size):
                try:
                    group.append(next(it))
                except StopIteration:
                    stopped = True
                    break
            if group:
                yield group


class Flatten(Nested):
    """Flatten the element of a dataset into multiple elements.

    The size of this dataset can't be determined in advance because we have to flatten every element from the wrapped
    dataset to know the exact size after flattening.
    """

    def __init__(self, dataset: Dataset, name: str = None):
        super().__init__(dataset, name)

    def generator(self):
        for chunk in self._dataset:
            for x in chunk:
                yield x


class Concat(NestedN, Sizable):
    """Concat multiple datasets.

    This dataset is a concatenation of multiple datasets.
    """

    def __init__(self, datasets: Iterable[Dataset], name: str = None):
        super().__init__(list(datasets), name)
        self._length = None

    @property
    def length(self):
        if self._length is None:
            self._length = sum(map(len, self._datasets))
        return self._length

    def generator(self):
        for x in itertools.chain(*self._datasets):
            yield x



"""
This module defines iterators to arrange a list of samples as a batch and process the samples to be fed into a model.
"""
from __future__ import annotations

import abc
import bisect
import collections
import itertools
import math
import warnings
from typing import *

import numpy

import lunas.dataset.core as core

__all__ = ['BatchIterator', 'ConstantIterator', 'BucketIterator', 'DataLoader', 'get_bucket_boundaries']

try:
    from torch.utils.data.dataset import IterableDataset as _IterableDataset
except ImportError:
    class _IterableDataset(object):
        ...


class BatchIterator(_IterableDataset, abc.ABC):

    def __init__(self, data_source: Union[core.Dataset, BatchIterator]) -> None:
        """Initialises the iterator.

        Args:
            data_source: an instance of Dataset or BatchIterator.
        """
        if not isinstance(data_source, (core.Dataset, BatchIterator)):
            raise TypeError(f'dataset ({type(data_source)}) must be instance of Dataset or BatchIterator.')

        self._data_source = data_source

    @abc.abstractmethod
    def generator(self) -> Iterator:
        raise NotImplementedError

    def state(self):
        return self._data_source.state()

    def load(self, state):
        self._data_source.load(state)

    def __iter__(self):
        num_workers = 1
        worker_id = 0

        try:
            import torch
            worker_info = torch.utils.data.get_worker_info()
            if worker_info is not None:
                num_workers = worker_info.num_workers
                worker_id = worker_info.id
        except ImportError:
            ...

        i = 0
        for samples in self.generator():
            if i == worker_id:
                yield samples
            i = (i + 1) % num_workers


class DataLoader(object):
    """Load and batch data.

    DataLoader wraps around a BatchIterator instance, batches multiple samples and optionally moves the data to pinned
    memory.

    This DataLoader relies on PyTorch's carefully designed `DataLoader` to implement safe multiprocessing operations.
    Since in a naive implementation, the main Python process can lose connection with child processes if one of them
    crashed, for example, by segment fault. However, PyTorch's DataLoader deals with this problem by tracking each
    process's states manually. Here we reuse it for simplicity.
    """

    def __init__(
            self,
            batch_iterator: BatchIterator,
            num_workers: int = 0,
            collate_fn: Callable[[List], Any] = None,
            pin_memory: bool = False,
            timeout: int = 0,
            worker_init_fn: Callable[[int], None] = None,
            multiprocessing_context=None
    ):
        if not isinstance(batch_iterator, BatchIterator):
            raise TypeError(f'batch_iterator ({type(batch_iterator)}) must be instance of BatchIterator.')

        self._batch_iterator = batch_iterator
        self._num_workers = num_workers
        self._collate_fn = collate_fn
        self._pin_memory = pin_memory
        self._timeout = timeout
        self._worker_init_fn = worker_init_fn
        self._multiprocessing_context = multiprocessing_context
        self._warning = False

        self._ptr = 0
        self._resumed = False

    def state(self):
        return {'ptr': self._ptr}

    def load(self, state):
        if not state:
            return
        self._ptr = state['ptr']
        self._resumed = True

    def _init_worker(self, worker_id):
        numpy.random.seed(numpy.random.get_state()[1][0])
        if self._worker_init_fn is not None:
            self._worker_init_fn(worker_id)

    def __iter__(self):
        if not self._resumed:
            self._ptr = 0
        self._resumed = False
        try:
            from torch.utils.data.dataloader import DataLoader as _DataLoader
            from torch.utils.data.dataset import IterableDataset as _IterableDataset
            dataloader = _DataLoader(self._batch_iterator, batch_size=None, num_workers=self._num_workers,
                                     collate_fn=self._collate_fn, pin_memory=self._pin_memory, timeout=self._timeout,
                                     worker_init_fn=self._init_worker,
                                     multiprocessing_context=self._multiprocessing_context)
            for batch in itertools.islice(dataloader, self._ptr, None):
                self._ptr += 1
                yield batch

        except ImportError:
            if not self._warning:
                self._warning = True
                warnings.warn(
                    'Multi-worker dataloader requires PyTorch installation, since PyTorch is not available, '
                    'main thread will be used for loading data.')

            for batch in itertools.islice(self._batch_iterator, self._ptr, None):
                if self._collate_fn is not None:
                    batch = self._collate_fn(batch)
                self._ptr += 1
                yield batch
        self._ptr = 0


class _Queue(object):

    def __init__(self) -> None:
        self._queue = collections.deque()

    def qsize(self):
        return len(self._queue)

    def empty(self):
        return self.qsize() == 0

    def push(self, x):
        self._queue.append(x)

    def pop(self):
        return self._queue.popleft()

    def pop_k(self, k=None):
        k = k or self.qsize()
        return [self.pop() for _ in range(k)]


def get_bucket_boundaries(inflation_rate: float = 1.1, inflation_offset: int = 0, min_length: int = 8,
                          max_length: int = 256, min_offset: int = 3):
    """
    Generate bucket boundaries to be used in BucketIterator. The value of inflation_rate and inflation_offset should
     be carefully picked to match the distribution of dataset. Greater value of inflation_rate or inflation_offset
     leads to sparse boundaries, which increases the chance of padding and thus reduces training efficiency;
     however, it also increases the diversity of samples in a batch and generate more diverse consecutive batches
     in terms of sample lengths. Sparse boundaries can avoid similar consecutive batches to some extent, which is
     bad for training.
    Args:
        inflation_rate:
        inflation_offset:
        min_length:
        max_length:
        min_offset:
    Returns:

    """
    if not (inflation_rate >= 1.0):
        raise ValueError(f'inflation_rate ({inflation_rate})  must be greater than 1.0.')
    if not (inflation_offset >= 0):
        raise ValueError(f'inflation_offset ({inflation_offset}) must be greater than 1.0.')

    x = min_length
    boundaries = []
    while x <= max_length:
        boundaries.append(x)
        x = max(x + min_offset, int(inflation_rate * x + inflation_offset))
    return boundaries


def _resize_batch_size(num_sample, required_batch_size_multiple):
    if num_sample <= required_batch_size_multiple:
        return num_sample
    return num_sample - (num_sample % required_batch_size_multiple)


class ConstantIterator(BatchIterator):
    """Constant-sized batch iterator.

    This iterator iterates over batches composed of a constant number of consecutive samples.

    """

    def __init__(self, data_source: Iterable, batch_size: int, drop_tail: bool = False) -> None:
        """Initialises the iterator.

        Args:
            data_source: a dataset as data source.
            batch_size: number of samples in a batch.
            drop_tail: whether to drop the trailing samples if they don't form a full batch.
        """
        super().__init__(data_source)
        self._batch_size = batch_size
        self._drop_tail = drop_tail
        self._num_batches = None

    def generator(self):
        it = iter(self._data_source)

        while True:
            samples = list(itertools.islice(it, self._batch_size))
            if len(samples) == self._batch_size:
                yield samples
            else:
                if not self._drop_tail and samples:
                    yield samples
                break

    def __len__(self):
        if self._num_batches is None:
            if self._drop_tail:
                self._num_batches = len(self._data_source) // self._batch_size
            else:
                self._num_batches = math.ceil(len(self._data_source) // self._batch_size)
        return self._num_batches


class BucketIterator(BatchIterator):
    """Varying-sized batch iterator.

    Varying-sized batch implies that the number of samples across different batches can differ. Since for each sample we can
    determine a particular size, for example, different text sequences have different number of tokens. In this regard,
    we define the batch size as the sum of sample sizes rather than the amount of samples.

    A range of buckets is created to hold similar sized samples into the same bucket for computational efficiency.
    Whenever a bucket reaches its capacity, it's yielded as a batch.
    """

    def __init__(
            self,
            data_source: Iterable,
            batch_size: int,
            get_length_fn: Callable[[Any], int],
            bucket_boundaries: List[int],
            min_length: int = 0,
            max_length: int = None,
            drop_tail: bool = False,
            required_batch_size_multiple: int = 1
    ) -> None:
        """Initialises the iterator.

        Args:
            data_source: a dataset object as data source.
            batch_size: batch size.
            get_length_fn: a function to evaluate the size of a given sample.
            bucket_boundaries: a list of ints that defines the buckets. Should be sorted in strictly ascending order.
            min_length: an int value indicates the minimum length of a sample, used as a filter.
            max_length: an int value indicates the maximum length of a sample, used as a filter.
            drop_tail: whether to drop the tail samples if they don't form a full batch.
            required_batch_size_multiple: adapts the actual batch size to be the multiple of this value.
            setting value to 8 to facilitate Tensor Core acceleration.

        """
        super().__init__(data_source)
        if not bucket_boundaries:
            raise ValueError(f'bucket_boundaries ({bucket_boundaries}) must be non-empty.')
        if not (bucket_boundaries == sorted(bucket_boundaries)):
            raise ValueError(f'bucket_boundaries ({bucket_boundaries}) must be in ascending order.')
        if not (len(set(bucket_boundaries)) == len(bucket_boundaries)):
            raise ValueError(f'bucket_boundaries ({bucket_boundaries}) must have unique elements.')
        if not callable(get_length_fn):
            raise ValueError(f'get_length_fn ({type(get_length_fn)}) must be callable.')
        if not (min_length <= bucket_boundaries[0]):
            raise ValueError(f'min_length ({min_length}) must be less than or equal to '
                             f'bucket_boundaries[0] ({bucket_boundaries[0]}).')
        if not (max_length is None or max_length >= bucket_boundaries[-1]):
            raise ValueError(f'max_length ({max_length}) must be less than or equal to '
                             f'bucket_boundaries[-1] ({bucket_boundaries[-1]}).')
        if not (required_batch_size_multiple > 0):
            raise ValueError(f'required_batch_size_multiple ({required_batch_size_multiple}) '
                             f'must be a positive integer.')

        max_length = max_length or bucket_boundaries[-1]

        self._get_length_fn = get_length_fn
        self._bucket_boundaries = bucket_boundaries
        self._batch_size = batch_size
        self._min_length = min_length
        self._max_length = max_length or bucket_boundaries[-1]
        self._drop_tail = drop_tail
        self._required_batch_size_multiple = required_batch_size_multiple

        if max_length > bucket_boundaries[-1]:
            bucket_boundaries.append(max_length)

        self._buckets = [_Queue() for _ in bucket_boundaries]
        self._capacities = [max(1, _resize_batch_size(batch_size // b, required_batch_size_multiple))
                            for b in bucket_boundaries]

    def generator(self):
        min_length = self._min_length
        max_length = self._max_length
        buckets = self._buckets
        capacities = self._capacities
        boundaries = self._bucket_boundaries
        num_bucket = len(buckets)
        len_fn = self._get_length_fn

        for x in self._data_source:
            size = len_fn(x)
            if size < min_length or size > max_length:
                continue
            i = bisect.bisect_left(boundaries, size)
            if i >= num_bucket:  # drop long sequence
                continue
            buckets[i].push(x)
            if buckets[i].qsize() == capacities[i]:
                yield buckets[i].pop_k(capacities[i])

        final_queue = _Queue()
        for q, capacity in zip(buckets, capacities):
            while not q.empty():
                final_queue.push(q.pop())
                if final_queue.qsize() >= capacity:
                    yield final_queue.pop_k(capacity)

        if not self._drop_tail and not final_queue.empty():
            yield final_queue.pop_k()

    def state(self):
        state = super().state()
        state['BucketIterator'] = {}
        state['BucketIterator']['_bucket_itr_queues'] = self._buckets
        state['BucketIterator']['_bucket_itr_capacities'] = self._capacities
        state['BucketIterator']['_bucket_itr_boundaries'] = self._bucket_boundaries
        return state

    def load(self, state):
        buckets = state['BucketIterator']['_bucket_itr_queues']
        capacities = state['BucketIterator']['_bucket_itr_capacities']
        bucket_boundaries = state['BucketIterator']['_bucket_itr_boundaries']
        if bucket_boundaries != self._bucket_boundaries:
            raise ValueError('Inconsistent bucket_boundaries.')
        if capacities != self._capacities:
            raise ValueError('Inconsistent bucket capacities.')
        self._buckets = buckets
        super().load(state)



import os
import sys

import lunas.dataset.core as core

__all__ = ['Stdin']


class Stdin(core.Dataset):
    """Stdin dataset.

    This dataset wraps `sys.stdin`.

    Never use this dataset in multiprocessing context in order not to observe unexpected behaviours since
    the correctness is not guaranteed.
    """

    def __init__(self, sentinel=None, name: str = None):
        super().__init__(name)
        if sentinel is None:
            sentinel = ''
        self._sentinel = sentinel + os.linesep
        self._resumable = False

    def generator(self):
        for x in sys.stdin:
            if self._sentinel == x:
                break
            yield x



import random
from typing import *

import lunas.dataset.core as core

__all__ = ['Sampling']


class Sampling(core.NestedN, core.Sizable):
    """Sampling dataset

    This dataset sample examples from multiple datasets by given weights.
    """

    def __init__(self, datasets: Iterable[core.Dataset], weights: Iterable[float] = None, virtual_size: int = None,
                 name: str = None):
        super().__init__(datasets, name)

        datasets = self._datasets

        if weights and sum(weights) != 1:
            raise ValueError(f'The sum of weights ({sum(weights)}) must be 1.0.')
        if virtual_size is not None and virtual_size < 0:
            raise ValueError(f'virtual_size ({virtual_size}) must be a non-negative value or None.')
        if weights:
            for ds, weight in zip(datasets, weights):
                if len(ds) == 0 and weight > 0:
                    raise ValueError(f'Attempt to sample from an empty dataset '
                                     f'with a non-zero weight ({weight}).')

        sizes = [len(ds) for ds in datasets]

        weights = weights if weights else [1.0 / len(datasets)] * len(datasets)

        max_weight_i, _ = max(enumerate(weights), key=lambda i_weight: i_weight[1])
        if virtual_size is None:
            virtual_size = int(sizes[max_weight_i] / weights[max_weight_i])
            virtual_size = sum(int(virtual_size * weight) for weight in weights)

        self._weights = weights
        self._virtual_size = virtual_size

    @property
    def length(self):
        return self._virtual_size

    def generator(self):
        indices = [i for i in range(len(self._datasets))]
        weights = [w for w in self._weights]
        datasets = [iter(dataset) for dataset in self._datasets]
        for _ in range(self._virtual_size):
            if len(indices) > 1:
                i = random.choices(indices, weights)[0]
            else:
                i = indices[0]
            try:
                yield next(datasets[i])
            except StopIteration:
                datasets[i] = iter(self._datasets[i])
                yield next(datasets[i])



import itertools
from pathlib import Path

import lunas.dataset.core as core

__all__ = ['TextLine']


def count_line(filename):
    f = open(filename, 'rb')
    buf_gen = itertools.takewhile(lambda x: x, (f.raw.read(1024 * 1024) for _ in itertools.repeat(None)))
    n = 0
    end_is_newline = True
    for buf in buf_gen:
        m = buf.count(b'\n')
        n += m
        if m > 0:
            end_is_newline = buf.rindex(b'\n') == len(buf) - 1
    n += int(not end_is_newline)
    return n


class TextLine(core.Dataset, core.Sizable):
    """TextLine dataset

    This dataset wraps a plain text file.
    """

    def __init__(self, filename: str, encoding: str = 'utf-8', name: str = None):
        super().__init__(name)
        filename = Path(filename)
        self._filename: Path = filename
        self._encoding = encoding
        self._length = count_line(filename)

    @property
    def length(self):
        return self._length

    def generator(self):
        with self._filename.open('r', encoding=self._encoding) as r:
            for line in r:
                yield line
